# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
import unittest

from wapiti.parsers import Decoder, Encoder

class TestJSONDecode(unittest.TestCase):

    def setUp(self):
        self.decoder = Decoder('json')

    def test_str(self):
        self.assertEqual(self.decoder.decode('"FOO"'), "FOO")
        self.assertRaises(TypeError, self.decoder.decode, ('FOO',))

    def test_list(self):
        self.assertEqual(self.decoder.decode('["FOO"]'), ["FOO"])
        
    def test_dict(self):
        self.assertEqual(self.decoder.decode('{"FOO":"BAR"}'), {"FOO": "BAR"})

    def test_nested(self):
        self.assertEqual(self.decoder.decode('{"FOO":["BAR"]}'), {"FOO": ["BAR"]})

class TestJSONEncode(unittest.TestCase):

    def setUp(self):
        self.encoder = Encoder('to_json')

    def test_str(self):
        self.assertEqual(self.encoder.encode("FOO"), '"FOO"')

    def test_list(self):
        self.assertEqual(self.encoder.encode(["FOO",]), '["FOO"]')
        
    def test_dict(self):
        self.assertEqual(self.encoder.encode({"FOO": "BAR"}), '{"FOO": "BAR"}')

    def test_nested(self):
        self.assertEqual(self.encoder.encode({"FOO":["BAR"]}), '{"FOO": ["BAR"]}')

if __name__ == '__main__':
    unittest.main()




# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'LimitTracking'
        db.create_table('wapiti_limittracking', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('key', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['wapiti.APIKey'])),
            ('limit', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['wapiti.Limit'])),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'], null=True)),
            ('session', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['sessions.Session'], null=True)),
            ('count', self.gf('django.db.models.fields.IntegerField')(default=1)),
            ('last_update', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
        ))
        db.send_create_signal('wapiti', ['LimitTracking'])

        # Adding model 'Limit'
        db.create_table('wapiti_limit', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('key', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['wapiti.APIKey'])),
            ('resource_regex', self.gf('django.db.models.fields.CharField')(max_length=256)),
            ('method', self.gf('django.db.models.fields.CharField')(max_length=8)),
            ('type', self.gf('django.db.models.fields.CharField')(max_length=8)),
            ('period', self.gf('django.db.models.fields.CharField')(max_length=8)),
            ('limit', self.gf('django.db.models.fields.IntegerField')()),
        ))
        db.send_create_signal('wapiti', ['Limit'])


    def backwards(self, orm):
        
        # Deleting model 'LimitTracking'
        db.delete_table('wapiti_limittracking')

        # Deleting model 'Limit'
        db.delete_table('wapiti_limit')


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'sessions.session': {
            'Meta': {'object_name': 'Session', 'db_table': "'django_session'"},
            'expire_date': ('django.db.models.fields.DateTimeField', [], {}),
            'session_data': ('django.db.models.fields.TextField', [], {}),
            'session_key': ('django.db.models.fields.CharField', [], {'max_length': '40', 'primary_key': 'True'})
        },
        'wapiti.apikey': {
            'Meta': {'object_name': 'APIKey'},
            'active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'default': "'5r7xl2TSgEhs1MVQqaYiRNAICWkc86Xm'", 'max_length': '32'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50', 'db_index': 'True'})
        },
        'wapiti.limit': {
            'Meta': {'object_name': 'Limit'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['wapiti.APIKey']"}),
            'limit': ('django.db.models.fields.IntegerField', [], {}),
            'method': ('django.db.models.fields.CharField', [], {'max_length': '8'}),
            'period': ('django.db.models.fields.CharField', [], {'max_length': '8'}),
            'resource_regex': ('django.db.models.fields.CharField', [], {'max_length': '256'}),
            'type': ('django.db.models.fields.CharField', [], {'max_length': '8'})
        },
        'wapiti.limittracking': {
            'Meta': {'object_name': 'LimitTracking'},
            'count': ('django.db.models.fields.IntegerField', [], {'default': '1'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['wapiti.APIKey']"}),
            'last_update': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'limit': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['wapiti.Limit']"}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['sessions.Session']", 'null': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True'})
        },
        'wapiti.permission': {
            'Meta': {'object_name': 'Permission'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['wapiti.APIKey']"}),
            'method': ('django.db.models.fields.CharField', [], {'max_length': '8'}),
            'resource_regex': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        }
    }

    complete_apps = ['wapiti']



# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
import base64
import cPickle as Pickle
import datetime as dt
import random
import re

from django.contrib.auth.models import User
from django.contrib.sessions.models import Session
from django.core.cache import cache
from django.db import models
from django.db.models import F

from wapiti.conf import APIKEY_ALLOWABLE_LETTERS, APIKEY_LENGTH

METHODS = ('GET', 'POST', 'PUT', 'DELETE')

def new_apikey():
    return ''.join(random.sample(APIKEY_ALLOWABLE_LETTERS, APIKEY_LENGTH))

class APIKey(models.Model):
    name = models.CharField(max_length=255)
    slug = models.SlugField()
    key = models.CharField(max_length=32, default=new_apikey)
    active = models.BooleanField(default=True)
    
    def __unicode__(self):
        return u'%s:%s (%s)'%(self.name, self.key, 
                              'act' if self.active else 'dis')

    def _perm_cache_key(self, method):
        return (self.key + '-' + method)

    def flush_permissions_cache(self):
        for m in METHODS:
            cache.delete(self._perm_cache_key(m))

    def is_authorized(self, request):
        q = self.permission_set.filter(
            method = request.method
        )

        if not q:
            return False

        perms_re = cache.get(self._perm_cache_key(request.method))

        if not perms_re:
            perms_re = re.compile(
                '(' 
                + ')|('.join(q.values_list('resource_regex', flat=True)) 
                + ')'
            )
            cache.set(self._perm_cache_key(request.method), perms_re)

        if perms_re.match(request.path):
            return True

        return False
    
    def check_limits(self, request):
        """
        Increments all limit usage counts and check them

        Returns exceeded limit object if any, True if none are exceeded.
        """
        for l in self.limit_set.filter(method=request.method):
            if l.increment_if_matches(self, request):
                return l
        return True

    def save(self, *args, **kwargs):
        # invalidate cached permission regexes
        self.flush_permissions_cache()
        super(APIKey, self).save(*args, **kwargs)

class Permission(models.Model):
    key = models.ForeignKey('APIKey')
    resource_regex = models.CharField(max_length=256)
    method = models.CharField(choices=zip(METHODS, METHODS), max_length=8)

    def __unicode__(self):
        return u'%s %s %s' % (
            self.key.name,
            self.method,
            self.resource_regex,
        )

    def save(self, *args, **kwargs):
        # invalidate cached permission regexes
        self.key.flush_permissions_cache()
        super(Permission, self).save(*args, **kwargs)

class Limit(models.Model):
    """
    A limit on #calls/time period for each user, session, or api key
    
    Note that limits are on an 'order deny, allow' basis: if you set
    multiple limits that all match a given query, if one of those limits
    is exceeded, the request 420's. So you can set a high blanket .* limit
    and then a much more restrictive per-call limit.
    """
    key = models.ForeignKey('APIKey', null=False)
    resource_regex = models.CharField(max_length=256, blank=False, null=False)
    method = models.CharField(choices=zip(METHODS, METHODS), max_length=8, 
                              null=False, blank=False)
    type = models.CharField(max_length=8, blank=False, null=False,
                            choices=(('session', 'per session'),
                                     ('user', 'per user'),
                                     ('key', 'per key'),
                                     ('ip', 'per ip')))
    period = models.CharField(max_length=8, blank=False, null=False,
                              choices=(('hour', 'per hour'),
                                       ('day', 'per day'),
                                       ('month', 'per month'),
                                       ('year', 'per year'),
                                       ('ever', 'forever')))
    limit = models.IntegerField(null=False)
    
    def __unicode__(self):
        self.key_str = self.key.name
        return ((u"%(limit)d per %(period)s per %(type)s "
                 u"on %(method)s:%(resource_regex)s for %(key_str)s"
                )%self.__dict__)
    
    @property
    def _resource_re(self):
        cache_key = u"limit%d"%self.id
        limit_re = cache.get(cache_key)
        if not limit_re:
            limit_re = re.compile(self.resource_regex)
            cache.set(cache_key, limit_re)
        return limit_re

    def increment_if_matches(self, apikey, request):
        """
        Increment counter if request matches this limit
        
        Returns True if limit is exceeded.
        """
        if ((request.method != self.method)
            or not self._resource_re.match(request.path)):
            return

        if self.type == 'session':
            querydict = {'session_id': request.session.session_key,
                         'key': apikey}
        elif self.type == 'ip':
            if 'REMOTE_ADDR' not in request.META:
                l.warning("Can't limit on IP - proxying server doesn't pass "
                          "REMOTE_ADDR - see wapiti README")
                return False
            querydict = {'ip': request.META['REMOTE_ADDR'],
                         'key': apikey}
        elif self.type == 'user':
            if request.user.is_anonymous():
                return True
            querydict = {'user': request.user.id, 'key': apikey}
        elif self.type == 'key':
            querydict = {'key': apikey}
        try:
            limit_count, created = self.limittracking_set.get_or_create(
                **querydict
            )
        except LimitTracking.MultipleObjectsReturned:
            # because many API requests might come in at the same time,
            # 2 django processes might call get_or_create at the same time,
            # and both will therefore create new objects
            # we take the highest count one, and delete the others
            limit_count = self.limittracking_set.filter(
                **querydict
            ).order_by('-count')[0]
            self.limittracking_set.exclude(id=limit_count.id).delete()

        return limit_count.increment()

class LimitTracking(models.Model):
    key = models.ForeignKey('APIKey', null=False)
    limit = models.ForeignKey("Limit", null=False)
    user = models.ForeignKey(User, null=True, blank=True)
    session_id = models.CharField(max_length=40, blank=True)
    count = models.IntegerField(default=0, null=False)
    ip = models.GenericIPAddressField(blank=True, null=True)
    last_update = models.DateTimeField(auto_now=True)

    def __unicode__(self):
        return u'%s out of %s' % (self.count, self.limit)

    def increment(self):
        """Resets counter based on time, or increments it

        Returns True if limit exceeded"""
        if self.limit.period != 'ever':
            now = getattr(dt.datetime.now(), self.limit.period)
            if now != getattr(self.last_update, self.limit.period):
                self.count = 1
                self.save()
                return self.count > self.limit.limit

        if self.count >= self.limit.limit:
            return True
        self.count = F('count') + 1
        self.save()

class LogItem(models.Model):
    """Log all API queries"""
    apikey = models.ForeignKey('APIKey', null=False)
    method = models.CharField(choices=zip(METHODS, METHODS), max_length=8, 
                              null=False, blank=False)
    resource = models.CharField(max_length=256, blank=False, null=False)
    pickled_arguments = models.TextField(blank=True)

    @classmethod
    def log_api_call(self, apikey, request, args):
        self.objects.create(
            apikey=apikey,
            method=request.method,
            resource=request.path,
            pickled_arguments=base64.encodestring(Pickle.dumps(args, 2))
        )

    @property
    def arguments(self):
        """Decode and unpickle the arguments back into python"""
        return Pickle.loads(base64.decodestring(self.pickled_arguments))





# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
import string
import re


APIKEY_ALLOWABLE_LETTERS = string.letters + string.digits
APIKEY_LENGTH = 32
ANONYMOUS_API_KEY = 'ANONYMOUS'
ID_RE = re.compile('[0-9]+')




# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.

from django.conf import settings

WAPITI_MAX_SLICE_SIZE = getattr(settings, 'WAPITI_MAX_SLICE_SIZE', 100)




# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
import datetime as dt
from decimal import Decimal
from inspect import getargspec
import json
import re
import warnings

from django.db import models
from django.db.models.fields.files import FieldFile
from django.db.models.query import QuerySet
from django.http import HttpResponse

from wapiti import helpers
from wapiti.modelapi import ModelApi

# ISO 8601
DATE_RE = re.compile('((?<!")[0-9]{4}-[0-1]?[0-9]-[0-3]?[0-9](?!"))')
DATE_FORMAT = '%Y-%m-%d'

class ModelNotRegisteredError(Exception):
    pass
 
HTML_STYLE = (
    """<style type="text/css">
    body {
        font-family: Helvetica;
        font-size: 9px;
    }

    table {
        width: 100%;
    }

    table tr td {
        vertical-align: top;
        padding: 4px;
        margin: 0px;
        font-weight: bold;
    }

    table tr td table {
        width: 900px;
        border: none;
        border-bottom: 1px solid #666;
    }
    table tr td table tr td table {
        width: 700px;
        border: none;
        border-top: 1px solid #aaa;
    }
    table tr td table tr td {
        font-weight: normal;
    }

    table tr td table tr td table tr td {
        border-bottom: 1px solid #aaa;
        font-weight: normal;
    }
    table tr td table tr td table tr td table {
        width: 500px;
        border: none;
    }

    table tr td table tr td table tr td table tr td table {
        width: 300px;
        border: none;
    }
    table tr td table tr td table tr td table tr td table tr td {
        border: none;
        font-weight: normal;
        padding: 0px;
        font-size: 0.8em;
    }

    </style>
    """
)
class EcoJSONDecoder(json.JSONDecoder):
    def decode(self, s):
        try:
            return super(EcoJSONDecoder, self).decode(s)
        except ValueError:
            return None

class Decoder(object):
    def __init__(self, format):
        self.format = format
        self.decode = getattr(self, format)

    def json(self, value):
        #FIXME detect dates and turn to string
        value = DATE_RE.subn(r'"\1"', value)[0]
        _decoded = json.loads(value, 'UTF-8', EcoJSONDecoder)
        _parsed = self.convert(_decoded)
        return _parsed
    
    def html(self, value):
        return self.json(value)
        
    def convert(self, value):
        # recursively decode objects and dates

        if isinstance(value, list):
            value = [self.convert(i) for i in value]
        elif isinstance(value, dict):
            for k, v in value.iteritems():
                value[k] = self.convert(v)
            if 'type' in value:
                try:
                    value = self.dict_to_object(value)
                except ModelNotRegisteredError:
                    # If we can't convert this dict to a model object,
                    # keep it as a dict.
                    pass
        elif isinstance(value, (str, unicode)) and DATE_RE.match(value):
            value = dt.datetime.strptime(value, DATE_FORMAT).date()
        return value

    def dict_to_object(self, value):
        try:
            m = helpers._registered_types[value['type']].api
        except KeyError:
            raise ModelNotRegisteredError()
        value.pop('type')
        if 'str' in value:
            value.pop('str')
        return m.objects.get(**value)

class Encoder(object):
    def __init__(self, format, jsonp=None, serialize_all_fields=False,
                 file_handler=None, max_depth=1):
        """
        Constructor, sets up Encoder behavior

        format: format to output (e.g. json)
        jsonp: in case the output needs to be wrapped in a jsonp method, give
               the method name here
        serialize_all_fields: by default, when encountering an object, the 
                              serializer will output a dict with the type, id,
                              repr() and only the fields in the ModelApi's
                              repr_object_fields list. Set this to true to 
                              override this and serialize all the fields
        file_handler: If set to a callable, the encoder will call this for 
                      outside code to do something with the file for filefields
                      and its descendants. The filefield will be passed as an argument.
        max_depth: when serialiazing fields pointing to other objects, the
                   serializer will recurse to at most max_depth depth
        """
        self.format = format
        self.encode = getattr(self, format)
        self.jsonp = jsonp
        self.serialize_all_fields = serialize_all_fields
        self.max_depth = max_depth
        self.file_handler = file_handler
        filehandler_args = getargspec(self.file_handler)[0]
        self._filehandler_backwards_compat = len(filehandler_args) > 1

    def to_json(self, value):
        return json.dumps(self.convert(value))

    def json(self, value):
        resp = self.to_json(value)
        if self.jsonp:
            jsonp = re.sub(r'[^a-zA-Z0-9_]', '', self.jsonp)
            resp = u'%s(%s)'%(jsonp, resp)
        return HttpResponse(resp, content_type='application/json; charset=utf-8')

    def html(self, value):
        converted = self.convert(value)
        return HttpResponse(
            '<html><head>%(style)s</head><body>%(body)s</body></html>' % {
                'style': HTML_STYLE,
                'body': self.to_html(converted)
            }, 
            content_type='text/html; charset=utf-8'
        )
    
    def to_html(self, value):
        if isinstance(value, (list, tuple, set)):
            if not len(value):
                html = '<em>none</em>'
            html = ('<table>'
                    + '\n'.join(['<tr><td>%s</td></tr>' % self.to_html(v)
                                 for v in value])
                    + '</table>')
        elif isinstance(value, dict):
            html = ('<table>'
                    + '\n'.join(['<tr><td>%s</td><td>%s</td></tr>' 
                                 % (self.to_html(k), self.to_html(v))
                                 for k, v in value.iteritems()])
                    + '</table>')
        elif value is None:
            html = u'<em>None</em>'
        else:
            html = (u'%s'%value).replace('\n', '<br/>')

        return html
        
    def convert(self, value, depth=1):
        # recursively encode objects and dates
        # the depth is used to monitor the recursion depth in the case 
        # of fields in a model which contain other model objects, such as
        # a FK. But because after the initial conversion, the encoder
        # will not encode all the object's fields even if serialize_all_fields
        # was True on the Encoder, this will only kick in if the ModelApi for 
        # that model had a FK/M2M/O2M field in its repr_object_fields

        if isinstance(value, (list, QuerySet, tuple, set)):
            value = [self.convert(i, depth) for i in value]
        elif isinstance(value, models.Model):
            if (depth == 1 and self.serialize_all_fields 
                and depth <= self.max_depth):
                value = self.convert(self.object_to_dict(value, all_fields=True),
                                     depth + 1)
            elif depth <= self.max_depth:
                value = self.convert(self.object_to_dict(value, all_fields=False),
                                     depth + 1)
            else:
                value = self.object_to_dict(value, all_fields=False)
        elif isinstance(value, dict):
            for k, v in value.iteritems():
                value[k] = self.convert(v, depth)
        elif isinstance(value, dt.date):
            value = value.strftime(DATE_FORMAT)
        elif isinstance(value, Decimal):
            value = float(value)
        elif isinstance(value, FieldFile):
            try:
                if callable(self.file_handler):
                    if self._filehandler_backwards_compat:
                        warnings.warn(
                            "3-arg form of file_handle is deprecated; "
                            "only the first argument is necessary",
                            DeprecationWarning)
                        self.file_handler(value, value.name, value.path)
                    else:
                        self.file_handler(value)
                value = {'file': value.name}
            except ValueError:
                value = {'file': None}
        elif hasattr(value, 'to_wapiti'):
            value = self.convert(value.to_wapiti(), depth)

        return value

    def object_to_dict(self, value, all_fields=False):
        for k, v in helpers._registered_types.iteritems():
            if isinstance(value, v.api.model):
                type_name = k
                api = v.api
                break
        else:
            type_name = value.__class__.__name__
            api = ModelApi()
        try:
            api_str = value.__api_unicode__()
        except AttributeError:
            api_str = unicode(value)

        obj_dict = {'type': type_name, 'id': value.pk, 'str': api_str}
        if all_fields:
            fields = [f.name for f in value._meta.fields if f.name not in api.invisible_fields]
        else:
            fields = api.object_repr_fields

        for f in fields:
            # object_repr_fields is referring to fields in a different model, which is fine
            # but leaving the dot in the field name sent is not nice on the other end,
            # particularly in python-land where you'll need to getattr everything
            key = f.replace('.', '__')
            obj_dict[key] = eval('value.' + f)
        return obj_dict




# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
# run with
# DJANGO_SETTINGS_MODULE=wapiti.settings python tests/__init__.py

from test_parsers import *

if __name__ == '__main__':
    unittest.main()




# encoding: utf-8
# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
"""API exception definitions for wapiti"""
import traceback

from django.conf import settings
from django.http import HttpResponse


class APIBaseException(Exception):
    name = "API Error"
    def __init__(self, msg='', code=500):
        self.msg, self.code = msg, code

    def __unicode__(self):
        return u"%d %s %s" % (self.code, self.name, self.msg)

    def get_resp(self):
        resp = HttpResponse(content=self.__unicode__(), status=self.code,
                            content_type='text/plain')
        resp.status_code = self.code
        return resp


class APINotFound(APIBaseException):
    name = "Not Found"
    def __init__(self, msg=''):
        super(APINotFound, self).__init__(msg, 404)

class APINotImplemented(APIBaseException):
    name = "Not Implemented"
    def __init__(self, msg=''):
        super(APINotImplemented, self).__init__(msg, 501)

class APIForbidden(APIBaseException):
    name = "Forbidden"
    def __init__(self, msg=''):
        super(APIForbidden, self).__init__(msg, 403)

class APIRateLimit(APIBaseException):
    name = "Rate Limited"
    def __init__(self, msg=''):
        super(APIRateLimit, self).__init__(msg, 420)

class APIMissingParameter(APIBaseException):
    name = "Missing Parameter(s)"
    def __init__(self, msg='', parameter='', all_parameters=()):
        super(APIMissingParameter, self).__init__(msg, 400)
        self.msg += "\nParameter missing: " + parameter
        if all_parameters:
            self.msg = (self.msg + "\nRequired parameters: "
                        + ' '.join(all_parameters))

class APIServerError(APIBaseException):
    name = "Server Error"
    def __init__(self, msg=''):
        super(APIServerError, self).__init__(msg, 500)
        if settings.DEBUG:
            self.msg += u"\nTraceback:\n " + traceback.format_exc().encode('utf-8')

class APIFormatNotSupported(APIBaseException):
    name = "Requested Format Not Supported"
    def __init__(self, msg='', format=''):
        super(APIFormatNotSupported, self).__init__(msg, 406)
        self.msg += " Format %s not in supported formats (%s)" % (
            format, ', '.join(SUPPORTED_FORMATS)
        )

class APICantGrokParameter(APIBaseException):
    name = "Can't Parse Parameter(s)"
    def __init__(self, msg='', parameter='', value=''):
        super(APICantGrokParameter, self).__init__(msg, 400)
        self.msg += " I can't decode parameter %s=%s" % (parameter, value)
        if settings.DEBUG:
            self.msg += "\nTraceback:\n " + traceback.format_exc()

    def __unicode__(self):
        return u"%d Can't grok: %s" % (self.code, self.msg)

class APIMethodNotAllowed(APIBaseException):
    name = "Method Not Allowed"
    def __init__(self, msg='', method='', allowed=()):
        super(APIMethodNotAllowed, self).__init__(msg, 405)
        self.msg = (self.msg + u" Method %s is not allowed." % method)
        if allowed:
            self.msg = self.msg + " Allowed methods are: " + ', '.join(allowed)

class APIBadSlice(APIBaseException):
    name = "Incorrect Slicing Parameters"
    def __init__(self, msg='', method='', allowed=()):
        super(APIBadSlice, self).__init__(msg, 416)
        self.msg = (self.msg + u" Bad slicing specification, or too much data requested.")

class APIPoorlyFormedQuery(APIBaseException):
    name = "Poorly Formed Query"

    def __init__(self, query_str='', msg=''):
        from wapiti.views.object_views import SearchView
        super(APIPoorlyFormedQuery, self).__init__(msg, 400)
        self.msg += (
            "\nMalformed query string (empty, or invalid json): "
            + query_str + '\n\n' + SearchView.__doc__
        )

class APIEvilQuery(APIBaseException):
    name = "Evil Query"

    def __init__(self, query_str='', msg=''):
        from wapiti.views.object_views import SearchView

        super(APIEvilQuery, self).__init__(msg, 400)
        self.msg += (
            "\nYour query is evil. Following keys is not permitted: "
            + query_str + '\n\n' + SearchView.__doc__
        )





# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
import inspect

from django.core.urlresolvers import reverse


from wapiti import helpers
from wapiti.exceptions import *
from wapiti.views.base_view import WapitiBaseView
from wapiti.views.object_views import SearchView, AutoCompleteView

class TopLevelInterfaceView(WapitiBaseView):
    def get(self, request):
        return APINotImplemented()

class InterfaceView(WapitiBaseView):
    def get(self, request, ver):
        types = {}
        for name, rt in helpers._registered_types.iteritems():
            methods = {}
            # FIXME: hardcode generic class methods
            search_url = reverse('search', kwargs={'ver' : ver, 'type' : name})
            methods['search'] = {
                'doc': '<pre>' + SearchView.__doc__ + '</pre>',
                'args': ('query', ),
                'type': 'class method',
                'url': search_url,
            }
            ac_url = reverse('auto_complete', 
                             kwargs={'ver' : ver, 'type' : name})
            methods['auto_complete'] = {
                'doc': '<pre>' + AutoCompleteView.__doc__ + '</pre>',
                'args': ('partial', ),
                'type': 'class method',
                'url': ac_url,
            }
            for m in inspect.getmembers(rt.api.model):
                if (inspect.ismethod(m[1])
                    and 'api' in dir(m[1])
                    and m[1].api):
                    if m[1].im_self:
                        # this is a class method
                        method_type = 'class method'
                        url = reverse('object_or_cm_view', 
                                      kwargs={'ver' : ver, 'type' : name, 
                                              'id_or_method' : m[0]})
                    else:
                        method_type = 'instance method'
                        url = reverse('im_view', 
                                      kwargs={'ver' : ver, 'type' : name, 
                                              'id' : 'OBJECT_ID',
                                              'method' : m[0]})

                    doc = (m[1].__doc__ and ('<pre>' + m[1].__doc__ + '</pre>')
                           or '<em>no documentation</em>'),
                    methods[m[0]] = {
                        'doc': doc,
                        'args': m[1].im_func.func_code.co_varnames[1:],
                        'type': method_type,
                        'url': url,
                    }
            types[name] = {
                'doc': rt.api.model.__doc__,
                'methods': methods,
            }
        return types




# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.

from django.conf.urls import url
from wapiti.views import (
    TopLevelInterfaceView, InterfaceView, AutoCompleteView, SearchView,
    ObjectOrClassMethodView, InstanceMethodView, Wapiti404View
)


urlpatterns = [
    url(r'^$', TopLevelInterfaceView.as_view(), name='top_level_interface'),
    url(r'^(?P<ver>[.0-9]+)/?$', InterfaceView.as_view(), name='interface'),
    url(r'^(?P<ver>[.0-9]+)/(?P<type>[a-zA-Z0-9_]+)/search/?$',
        SearchView.as_view(), name='search'),
    url(r'^(?P<ver>[.0-9]+)/(?P<type>[a-zA-Z0-9_]+)/auto_complete/?$',
        AutoCompleteView.as_view(), name='auto_complete'),
    url(r'^(?P<ver>[.0-9]+)/(?P<type>[a-zA-Z0-9_]+)'
        r'/(?P<id_or_method>[a-zA-Z0-9_]+)/?$',
        ObjectOrClassMethodView.as_view(), name='object_or_cm_view'),
    url(r'^(?P<ver>[.0-9]+)/(?P<type>[a-zA-Z0-9_]+)/'
        r'(?P<id>[a-zA-Z0-9_]+)/(?P<method>[a-zA-Z0-9_]+)/?$',
        InstanceMethodView.as_view(), name='im_view'),
    url(r'.*', Wapiti404View.as_view(), name='404'),
]



# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
from django.contrib import admin

from models import *

class PermissionInline(admin.TabularInline):
    model = Permission

class LimitInline(admin.TabularInline):
    model = Limit

class APIKeyAdmin(admin.ModelAdmin):
    prepopulated_fields = {"slug": ("name",)}
    inlines = [
        PermissionInline,
        LimitInline,
    ]

class PermissionAdmin(admin.ModelAdmin):
    pass

class LimitTrackingInline(admin.TabularInline):
    model = LimitTracking

class LimitAdmin(admin.ModelAdmin):
    inlines = [
        LimitTrackingInline,
    ]

class LimitTrackingAdmin(admin.ModelAdmin):
    model = LimitTracking
    list_display = ('count', 'user', 'session_id', 'last_update', 'limit')

admin.site.register(APIKey, APIKeyAdmin)
admin.site.register(Permission, PermissionAdmin)
admin.site.register(Limit, LimitAdmin)
admin.site.register(LimitTracking, LimitTrackingAdmin)




# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
from collections import namedtuple
from decorator import decorator
import sys

from django.apps import apps

from wapiti.conf import ID_RE

_RegisteredType = namedtuple('RegisteredType', ('api', ))

_registered_types = {}


def register(name, modelapi):
    """Register a model with the API"""
    global _registered_types
    if len(sys.argv) > 1 and sys.argv[1] in ('syncdb', 'migrate'):
        return
    if modelapi.__name__ in _registered_types:
        return

    api = modelapi()
    api.objects = getattr(api, 'objects', api.model.objects)
    api.search_objects = getattr(api, 'search_objects', api.objects)
    _registered_types[name] = _RegisteredType(api=api)


def _api_method(f, *args, **kwargs):
    return f(*args, **kwargs)


def api_method(f):
    """Decorator to declare a method api-accessible"""
    f.api = True
    return decorator(_api_method, f)


def _is_id(id):
    return ID_RE.match(id)


def _register_models():
    """Find app api submodules and register models"""
    for app in apps.get_app_configs():
        try:
            __import__(
                '{}.{}'.format(app.name, 'api'), globals(), locals()
            )
        except ImportError:
            pass


def register_models():
    _register_models()



# encoding: utf-8
# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
from collections import defaultdict
import copy
import json
import inspect
import re
import sys

from django import http
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models.query import QuerySet
from django.http import HttpResponse
from django.template import RequestContext, loader
from django.utils.translation import ugettext_lazy as _
from functools import update_wrapper

from wapiti import helpers
from wapiti.conf import ANONYMOUS_API_KEY
from wapiti.exceptions import *
from wapiti.models import APIKey, LogItem
from wapiti.parsers import Decoder, Encoder
import wapiti.settings as wsettings

SUPPORTED_FORMATS = ('json', 'html')

class classonlymethod(classmethod):
    def __get__(self, instance, owner):
        if instance is not None:
            raise AttributeError("This method is available only on the view class.")
        return super(classonlymethod, self).__get__(instance, owner)

class View(object):
    """
    Intentionally simple parent class for all views. Only implements
    dispatch-by-method and simple sanity checking.
    """
    # adapted from django 1.3's base class view

    http_method_names = ['get', 'post', 'put', 'delete', 'head', 'options', 
                         'trace']

    def __init__(self, **kwargs):
        """
        Constructor. Called in the URLconf; can contain helpful extra
        keyword arguments, and other things.
        """
        # Go through keyword arguments, and either save their values to our
        # instance, or raise an error.
        for key, value in kwargs.iteritems():
            setattr(self, key, value)

    @classonlymethod
    def as_view(cls, **initkwargs):
        """
        Main entry point for a request-response process.
        """
        # sanitize keyword arguments
        for key in initkwargs:
            if key in cls.http_method_names:
                raise TypeError(u"You tried to pass in the %s method name as a "
                                u"keyword argument to %s(). Don't do that."
                                % (key, cls.__name__))
            if not hasattr(cls, key):
                raise TypeError(u"%s() received an invalid keyword %r" % (
                    cls.__name__, key))

        def view(request, *args, **kwargs):
            self = cls(**initkwargs)
            return self.dispatch(request, *args, **kwargs)

        # take name and docstring from class
        update_wrapper(view, cls, updated=())

        # and possible attributes set by decorators
        # like csrf_exempt from dispatch
        update_wrapper(view, cls.dispatch, assigned=())
        return view

    def dispatch(self, request, *args, **kwargs):
        # Try to dispatch to the right method; if a method doesn't exist,
        # defer to the error handler. Also defer to the error handler if the
        # request method isn't on the approved list.
        if request.method.lower() in self.http_method_names:
            handler = getattr(self, request.method.lower(), 
                              self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed
        self.request = request
        return handler(request, *args, **kwargs)

    def http_method_not_allowed(self, request, *args, **kwargs):
        allowed_methods = [m for m in self.http_method_names 
                           if hasattr(self, m)]
        return APIMethodNotAllowed(method=request.method, 
                                   allowed=allowed_methods).get_resp()

class WapitiBaseView(View):
    def dispatch(self, request, *args, **kwargs):
        resp = self._dispatch(request, *args, **kwargs)

        if isinstance(resp, HttpResponse):
            return resp
        elif isinstance(resp, APIBaseException):
            resp = resp.get_resp()
        else:
            if isinstance(resp, (list, tuple, QuerySet)):
                resp = resp[self.slice_left:self.slice_right]
            try:
                resp = Encoder(self.format, jsonp=self.jsonp,
                               serialize_all_fields=self.all_fields,
                               max_depth=self.depth).encode(resp)
            except:
                return APIServerError(u"Error encoding the results!").get_resp()

        return resp

    def _dispatch(self, request, *args, **kwargs):
        # always check API Key permissions
        self.args = defaultdict(lambda: '')
        for k, v in request.GET.iteritems():
            self.args[k] = v
        for k, v in request.POST.iteritems():
            self.args[k] = v

        self.format = self.args.pop('format', 'json')
        self.jsonp = self.args.pop('callback', None)
        self.all_fields = self.args.pop('all_fields', False)
        self.depth = min(self.args.pop('depth', 1), 5)
        if self.format not in SUPPORTED_FORMATS:
            return APIFormatNotSupported(format=self.format)
        
        self.api_key = self.args.pop('k', ANONYMOUS_API_KEY)

        authorized = True
        try:
            apikey = APIKey.objects.get(key=self.api_key, active=True)
        except APIKey.DoesNotExist:
            authorized = False
        else:
            authorized = apikey.is_authorized(request)

        if authorized:
            lim = apikey.check_limits(request)
            if lim != True:
                return APIRateLimit("Limit exceeded: %s"%lim)
        else:
            return APIForbidden("Invalid API Key")

        # parse slicing arguments, if any
        self.slice_left = self.args.pop('slice_left', 0)
        self.slice_right = self.args.pop('slice_right', 100)
        try:
            self.slice_left = int(self.slice_left)
            self.slice_right = int(self.slice_right)
        except ValueError:
            return APIBadSlice()

        if (self.slice_left >= self.slice_right
            or self.slice_left < 0 or self.slice_right < 1
            or (self.slice_right - self.slice_left) > wsettings.WAPITI_MAX_SLICE_SIZE):

            return APIBadSlice()

        # parse the arguments        
        self._decoder = Decoder(self.format)
        for k, v in self.args.iteritems():
            try:
                self.args[k] = self._decoder.decode(v)
            except:
                return APICantGrokParameter(k, v)

        try:
            resp = super(WapitiBaseView, self).dispatch(request, *args, 
                                                        **kwargs)
        except APIBaseException, e:
            return e
        LogItem.log_api_call(apikey, request, dict(self.args))

        return resp

class Wapiti404View(View):
    def get(self, request):
        return APINotFound().get_resp()

    def put(self, request):
        return APINotFound().get_resp()

    def delete(self, request):
        return APINotFound().get_resp()

    def post(self, request):
        return APINotFound().get_resp()





# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'APIKey'
        db.create_table('wapiti_apikey', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('slug', self.gf('django.db.models.fields.SlugField')(max_length=50, db_index=True)),
            ('key', self.gf('django.db.models.fields.CharField')(default='0hHA43FivUGPJz8CsnKqrLeNXwbYM5TW', max_length=32)),
            ('active', self.gf('django.db.models.fields.BooleanField')(default=True)),
        ))
        db.send_create_signal('wapiti', ['APIKey'])

        # Adding model 'Permission'
        db.create_table('wapiti_permission', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('key', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['wapiti.APIKey'])),
            ('resource_regex', self.gf('django.db.models.fields.CharField')(max_length=256)),
            ('method', self.gf('django.db.models.fields.CharField')(max_length=8)),
        ))
        db.send_create_signal('wapiti', ['Permission'])


    def backwards(self, orm):
        
        # Deleting model 'APIKey'
        db.delete_table('wapiti_apikey')

        # Deleting model 'Permission'
        db.delete_table('wapiti_permission')


    models = {
        'wapiti.apikey': {
            'Meta': {'object_name': 'APIKey'},
            'active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'default': "'iKByN40ZlzhGEMIJOqvCwctDTYrHsm87'", 'max_length': '32'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50', 'db_index': 'True'})
        },
        'wapiti.permission': {
            'Meta': {'object_name': 'Permission'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['wapiti.APIKey']"}),
            'method': ('django.db.models.fields.CharField', [], {'max_length': '8'}),
            'resource_regex': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        }
    }

    complete_apps = ['wapiti']



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings
import wapiti.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='APIKey',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('slug', models.SlugField()),
                ('key', models.CharField(default=wapiti.models.new_apikey, max_length=32)),
                ('active', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='Limit',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('resource_regex', models.CharField(max_length=256)),
                ('method', models.CharField(max_length=8, choices=[(b'GET', b'GET'), (b'POST', b'POST'), (b'PUT', b'PUT'), (b'DELETE', b'DELETE')])),
                ('type', models.CharField(max_length=8, choices=[(b'session', b'per session'), (b'user', b'per user'), (b'key', b'per key'), (b'ip', b'per ip')])),
                ('period', models.CharField(max_length=8, choices=[(b'hour', b'per hour'), (b'day', b'per day'), (b'month', b'per month'), (b'year', b'per year'), (b'ever', b'forever')])),
                ('limit', models.IntegerField()),
                ('key', models.ForeignKey(to='wapiti.APIKey')),
            ],
        ),
        migrations.CreateModel(
            name='LimitTracking',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('session_id', models.CharField(max_length=40, blank=True)),
                ('count', models.IntegerField(default=0)),
                ('ip', models.IPAddressField(null=True, blank=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('key', models.ForeignKey(to='wapiti.APIKey')),
                ('limit', models.ForeignKey(to='wapiti.Limit')),
                ('user', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='LogItem',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('method', models.CharField(max_length=8, choices=[(b'GET', b'GET'), (b'POST', b'POST'), (b'PUT', b'PUT'), (b'DELETE', b'DELETE')])),
                ('resource', models.CharField(max_length=256)),
                ('pickled_arguments', models.TextField(blank=True)),
                ('apikey', models.ForeignKey(to='wapiti.APIKey')),
            ],
        ),
        migrations.CreateModel(
            name='Permission',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('resource_regex', models.CharField(max_length=256)),
                ('method', models.CharField(max_length=8, choices=[(b'GET', b'GET'), (b'POST', b'POST'), (b'PUT', b'PUT'), (b'DELETE', b'DELETE')])),
                ('key', models.ForeignKey(to='wapiti.APIKey')),
            ],
        ),
    ]



# encoding: utf-8
# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
import os
from setuptools import setup, find_packages

description = "Django API layer"
try:
    cur_dir = os.path.dirname(__file__)
    long_description = open(os.path.join(cur_dir, 'README.rst')).read()
except:
    long_description = description

setup(
    name = "django-wapiti",
    version = "1.0.1",
    packages = find_packages(),
    description = description,
    author = "Ecometrica",
    author_email = "dev@ecometrica.com",
    maintainer = "Rory Geoghegan",
    maintainer_email = "rory.geoghegan@ecometrica.com",
    url = "http://github.com/ecometrica/django-wapiti/",
    keywords = ["django", "api"],
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Topic :: Software Development :: Internationalization",
        "Framework :: Django",
        ],
    long_description = long_description,
)



# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.
class ModelApi(object):
    # mode this api type is tied to, if any
    model = None
    # fields that should always be returned when serializing the object
    object_repr_fields = []
    # fields to search into for auto_complete
    auto_complete_fields = []
    # auto_complete lookup type
    auto_complete_type = 'istartswith'
    # order auto_complete and search results by
    # this can be 'LENGTH:fieldname' in which case results will be sorted
    # by the length of field fieldname (which should be a string!)
    order_by = 'id'
    # never ever allow modification of these fields, whoever asks
    read_only_fields = []
    # never ever show these fields, whoever asks
    invisible_fields = []
    # fields which can be traversed in search. e.g. to be able to do:
    # ['foofield__name','icontains','BOO']
    # you'll need to put 'foofield' in here
    traversable_fields = []
    # the queryset that will be used for all direct accesses to these objects,
    # such as object views or calling an instance method
    # you could create a custom queryset class to allow for an api
    # type that doesn't map to an actual django type, or simply use this
    # to filter objects available through the api
    # left out so you can hook it up lazily
    # objects = None

    # queryset/manager to use when searching; this allows you to setup a more restricted
    # set of objects when searching, while objects above would be wider, and allow older
    # references to still be valid when used directly
    # if not overridden, the above objects attribute will be used for searches too
    #search_objects = None
    



# Copyright (c) Ecometrica. All rights reserved.
# Distributed under the BSD license. See LICENSE for details.

from django.db.models import Q

from wapiti import helpers
from wapiti.views.base_view import WapitiBaseView
from wapiti.exceptions import *

class WapitiTypeBaseView(WapitiBaseView):
    def dispatch(self, request, ver, type, *args, **kwargs):
        # check if type is registered barf if not
        try:
            self.model = helpers._registered_types[type].api.model
            self.api = helpers._registered_types[type].api
        except KeyError:
            return APINotFound().get_resp()
        return super(WapitiTypeBaseView, self).dispatch(request, ver, type, 
                                                        *args, **kwargs)

class ObjectOrClassMethodView(WapitiTypeBaseView):
    def get(self, request, ver, type, id_or_method, *args, **kwargs):
        
        # determine if id_or_method is an id, call object_view if so
        if helpers._is_id(id_or_method):
            return self._object_view(request, type, id_or_method)
        else:
            # otherwise call class_method
            return self._class_method(request, type, id_or_method)

    def _object_view(self, request, type, id):
        try:
            return self.api.objects.get(id=id)
        except (self.model.DoesNotExist, ValueError):
            return APINotFound("%s with id %s not found" % (type, id))

    def _class_method(self, request, type, method):

        # check if method exists
        try:
            # note: this looks dangerous - the type and method are passed from the
            # client - but the urls definition prevents any other char than
            # [a-zA-Z_]
            m = eval('self.model.' + method)
        except AttributeError:
            return APINotFound("No method %s on type %s" % (method, type))

        # check if method is registered with the API
        try:
            if not m.api:
                return APIForbidden("Method not available through API")
        except AttributeError:
            return APIForbidden("Method not available through API")
        
        # check if method is a class method
        if m.im_self is not self.model:
            return APINotFound("Method is an instance method")

        return m(**self.args)

def _parse_order_by(results, order_by):
    if order_by.startswith('LENGTH:'):
        field_name = order_by.split(':')[1]
        results = results.extra(
            select={'FIELD_LENGTH':'Length(%s)'%field_name})
        order_by = 'FIELD_LENGTH'
    else:
        order_by = order_by
    return results.order_by(order_by)

class SearchView(WapitiTypeBaseView):
    """
    Search objects of this type on arbitrarily complex queries

    Query is a list of lists and operators, each list containing the field to 
    search for, the search type and the search term. Example queries:
        [["name", "iexact", "Long-haul, business"]]
        [["name", "iexact", "Long-haul, business"], 
         or, ["name", "icontains", "short-haul"]]
    Supported search types: 
        "equal"/"iexact": case-sensitive/insensitive full string match 
                          or "equal" for numerical equality, 
        "contains"/"icontains": case-sensitive/insensitive substring match, 
        "regex"/"iregex": regular expression case-sensitive/insensitive 
                          string match, 
        "lt"/"gt": less-than, greater-than numerical search
    You can't follow foreign or reverse keys.
    """
        
    _CONDITIONS = {'not': '__invert__', 'and': '__and__', 'or': '__or__'}

    def get(self, request, ver, type, *args, **kwargs):
        
        return self._search(request, type)
    
    def _search(self, request, type):
        # parse search query
        query = self.args['query']
        if not query:
            raise APIPoorlyFormedQuery('')
        query_q = self._parse_search_query(query)
        search_results = self.api.search_objects.filter(query_q)
        search_results = _parse_order_by(search_results, self.api.order_by)
        search_results = search_results.distinct()
        return search_results


    def _parse_search_query(self, query):
        if not query:
            return Q()

        elif (isinstance(query, list) and len(query) == 3 
              and isinstance(query[0], (str, unicode))
              and isinstance(query[1], (str, unicode))):
            # this is a query atom, such as ["foo","contains","bar"]
            if ('__' in query[0] 
                and query[0][:query[0].rfind('__')] 
                not in self.api.traversable_fields):

                    raise APIEvilQuery(
                        "You tried following %s but only %s are traversable."%(
                            query[0].split('__')[0], ', '.join(self.api.traversable_fields)
                        )
                    )
            return Q(**{'__'.join(query[:2]) : query[2]})

        elif isinstance(query, list):
            # query contains atoms and conditions
            q = None
            cond = '__or__'
            for t in query:
                if isinstance(t, list):
                    new_q = self._parse_search_query(t)
                    if q is None:
                        q = new_q
                    else:
                        q = getattr(q, cond)(new_q)
                else:
                    cond = self._CONDITIONS[t.lower()]
            return q


class AutoCompleteView(WapitiTypeBaseView):
    """
    Search suitable for auto-completion of user-entered partial strings

    Partial is a string to be searched for in a case-insensitve fashion. 
    The fields that will be searched are defined per-type.
    """
    def get(self, request, ver, type, *args, **kwargs):
        
        return self._auto_complete(request, type)
    
    def _auto_complete(self, request, type):
        search_str = self.args['partial']
        if not search_str:
            return []

        
        search_q = Q()
        for f in self.api.auto_complete_fields:
            search_d = {'%s__%s' % (f, self.api.auto_complete_type): search_str}
            search_q |= Q(**search_d)
        results = self.api.search_objects.filter(search_q)
        results = _parse_order_by(results, self.api.order_by)
        results = results.distinct()
        return list(results)



class InstanceMethodView(WapitiTypeBaseView):
    def get(self, request, ver, type, id, method):

        # check if object exists
        try:
            self.obj = self.api.objects.get(id=id)
        except self.model.DoesNotExist:
            return APINotFound("No such object")
        except ValueError:
            return APINotFound("No such object: malformed id")

        # check if method exists
        try:
            # note: this looks dangerous - the type and method are passed from the
            # client - but the urls definition prevents any other char than
            # [a-zA-Z_]
            self.method = eval('self.obj.' + method)
        except AttributeError:
            return APINotFound("No such method available through the API")

        # check if method is registered with the API
        try:
            if not self.method.api:
                return APIForbidden("Method not available through API")
        except AttributeError:
            return APIForbidden("Method not available through API")

        # check if method is an instance method
        if self.method.im_self is self.model:
            return APINotFound("Method is a class method")

        return self.method(**self.args)

